/*
 * PartitionPolicyImpl.hh
 *
 * Copyright (c) 2003, 2004 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file PartitionPolicyImpl.hh
 *
 * Header file for the PartitionPolicyImpl class.
 */

#ifndef _partition_policy_impl_hh
#define _partition_policy_impl_hh

#include "BrokerC.h"
#include "BrokerS.h"

#include "StrictPolicyS.h"

#include "listNode.h"

/**
 * An implementation of the PartitionPolicy interface.
 */
class PartitionPolicyImpl : public virtual POA_BrokerPolicies::PartitionPolicy
{

public:

    /**
     * Construct a PartitionPolicyImpl object with the given values.
     *
     * @param name The name of this object as registered with the
     * NamingService.
     */
    PartitionPolicyImpl(const char *name);

    /**
     * Deconstruct the policy.
     */
    virtual ~PartitionPolicyImpl(void);

    /** @copydoc Broker::Policy::Name */
    virtual char *Name(void)
	throw (CORBA::SystemException);
    
    /** @copydoc Broker::Policy::AddTask */
    virtual void AddTask(Broker::Task_ptr new_task,
			 const Broker::ScheduleParameters &sp)
	throw (CORBA::SystemException,
	       Broker::DuplicateScheduleParameter,
	       Broker::InvalidScheduleParameter,
	       Broker::MissingScheduleParameter);
    /** @copydoc Broker::Policy::RemoveTask */
    virtual void RemoveTask(Broker::Task_ptr added_task)
	throw (CORBA::SystemException);
    
    /** @copydoc Broker::Policy::GetTaskList */
    virtual Broker::TaskList *GetTaskList(void)
	throw (CORBA::SystemException);

    /** @copydoc Broker::Policy::Activate */
    virtual void Activate(const Broker::TaskList &tasks)
	throw (CORBA::SystemException);
    /** @copydoc Broker::Policy::Deactivate */
    virtual void Deactivate(void)
	throw (CORBA::SystemException);
    
    /** @copydoc Broker::Policy::ChangeTaskCPU */
    virtual Broker::CPUReserve ChangeTaskCPU(Broker::RealTimeTask_ptr task,
					     const Broker::CPUReserve &advice)
	throw (CORBA::SystemException, Broker::InvalidState);

    /** @copydoc Broker::Policy::GetMaxCPUAllocation */
    virtual CORBA::Float GetMaxCPUAllocation(void)
	throw (CORBA::SystemException);
    /** @copydoc Broker::Policy::SetMaxCPUAllocation */
    virtual void SetMaxCPUAllocation(CORBA::Float amount)
	throw (CORBA::SystemException);

    /** @copydoc BrokerPolicies::PartitionPolicy::DefaultPolicy */
    virtual void DefaultPolicy(Broker::Policy_ptr policy)
	throw (CORBA::SystemException);

    /** @copydoc BrokerPolicies::PartitionPolicy::DefaultPolicy */
    virtual Broker::Policy_ptr DefaultPolicy(void)
	throw (CORBA::SystemException);

    /** @copydoc BrokerPolicies::PartitionPolicy::SetTaskPolicy */
    virtual void SetTaskPolicy(Broker::Task_ptr task,
			       Broker::Policy_ptr bp)
	throw (CORBA::SystemException);

    /** @copydoc BrokerPolicies::PartitionPolicy::GetTaskPolicy */
    virtual Broker::Policy_ptr GetTaskPolicy(Broker::Task_ptr task)
	throw (CORBA::SystemException);

private:

    /**
     * List node used to track tasks that this policy is managing.
     */
    struct TaskData {
	struct lnMinNode td_Link;
	Broker::RealTimeTask_var td_Task;
	Broker::ScheduleParameters_var td_ScheduleParameters;
	Broker::Policy_var td_Policy;
    };

    /**
     * Repair the task list by removing any dead/unreachable objects.
     */
    void RepairTaskList(void);

    /**
     * Find the TaskData node that corresponds to the given task.
     *
     * @param task The Broker::RealTimeTask object to look up.
     * @return The TaskData list node that has the given task object.
     * @exception Broker::Internal If no TaskData node exists for the given
     *            task.
     */
    struct TaskData *FindTaskData(Broker::RealTimeTask_ptr task);
    
    /** The maximum percentage of CPU time that we are allowed to allocate. */
    float pp_MaxUsedCPU;

    /** The name of this object as registered with the NamingService. */
    CORBA::String_var pp_Name;

    /** The default subpolicy to use for newly added tasks. */
    Broker::Policy_var pp_DefaultPolicy;

    /** A list of TaskData objects */
    struct lnMinList pp_List;
    
};

#endif
