/*
 * rt_scheduler_test.c
 *
 * Copyright (c) 2004 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file rt_scheduler_test.c
 *
 * Test cases for the TimeSys implementation of the rt_scheduler interface.
 */

#include "config.h"

#ifndef DEBUG
#define DEBUG
#endif

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

#include <assert_pp.h>

#include <rk/rk.h>
#include "rk_util.h"
#include "rt_scheduler.h"

int main(int argc, char *argv[])
{
    int retval = EXIT_SUCCESS;

    /**
     * @test The rts_query_scheduler function.
     */
    {
	struct rts_scheduler_data rsd;

	ensure(rts_query_scheduler(&rsd) == 0);
	ensure(rsd.rsd_Flags & RSDF_RESERVATIONS);
	ensure(rsd.rsd_MaxReservable > 0.0);
    }

    /**
     * @test Creating an rts_schedulable for the current task.
     */
    {
	rts_schedulable_t rs;

	ensure(rk_proc_get_rset(getpid()) == NULL_RESOURCE_SET);
	ensure(rts_grab_schedulable(&rs, getpid(), NULL) == 0);
	ensure(rk_proc_get_rset(getpid()) != NULL_RESOURCE_SET);
	ensure(rts_release_schedulable(rs) == 0);
	ensure(rk_proc_get_rset(getpid()) == NULL_RESOURCE_SET);
    }

    /**
     * @test Creating an rts_schedulable, with a specific name, for the current
     * task.
     */
    {
	static const char *this_name = "rts_test";
	char rs_name[RK_NAME_LEN + 1];
	rk_resource_set_t this_rs;
	rts_schedulable_t rs;

	ensure(rk_proc_get_rset(getpid()) == NULL_RESOURCE_SET);
	ensure(rts_grab_schedulable(&rs, getpid(), this_name) == 0);
	ensure((this_rs = rk_proc_get_rset(getpid())) != NULL_RESOURCE_SET);
	ensure(rk_resource_set_get_name(this_rs, rs_name) == 0);
	ensure(strcmp(rs_name, this_name) == 0);
	ensure(rts_release_schedulable(rs) == 0);
	ensure(rk_proc_get_rset(getpid()) == NULL_RESOURCE_SET);
    }

    /**
     * @test Creating two rts_schedulables and making sure the resource set is
     * not released until the first one is released.
     */
    {
	rts_schedulable_t rs1, rs2;

	ensure(rk_proc_get_rset(getpid()) == NULL_RESOURCE_SET);
	ensure(rts_grab_schedulable(&rs1, getpid(), NULL) == 0);
	ensure(rk_proc_get_rset(getpid()) != NULL_RESOURCE_SET);
	ensure(rts_grab_schedulable(&rs2, getpid(), NULL) == 0);
	ensure(rts_release_schedulable(rs2) == 0);
	ensure(rk_proc_get_rset(getpid()) != NULL_RESOURCE_SET);
	ensure(rts_release_schedulable(rs1) == 0);
	ensure(rk_proc_get_rset(getpid()) == NULL_RESOURCE_SET);
    }
    
    return( retval );
}
